/**************************************************************/
/* clsieve.c                                                  */
/* Chris Monico, 1/7/03                                       */
/* Classical sieve for the NFS.                               */
/**************************************************************/
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ggnfs.h"


#define SKIP_R_PRIMES 0
#define SKIP_A_PRIMES 0

#define VERY_SMALL_ENTRY -127
#define VERY_SMALL_SHORT -32767


/* This function should be changed in one of two ways, to remove the `if (evenB)'
   from within the loop. Either:
   (1) Make a seperate loop or function for the b-even case.
   (2) Add a new variable which is set outside the loop (probably to either 1 or 2,
       depending on whether or not b is even) and a small calculation inside the
       loop which will work in either case, but avoid the condition-testing.
*/

/*******************************************************************************/
int clSieve(s32 *smooth_a, int maxSmooth, nfs_fb_t *FB, s32 _a0, s32 _a1, s32 b, 
            s32 cacheSize)
/*******************************************************************************/
/* Sieve from (a0, b) to (a1, b). Put the smooth a's in smooth_a.              */
/* That is, sieve on a - b*m, and store the values of 'a' that                 */
/* give smooth values.                                                         */
/* Return value: Number of relations found, or < 0 on error.                   */
/*******************************************************************************/
{ static int    initialized=0;
  static mpz_t  bm, t;
  static short  *tmp;
  short register *t2, *tStop;
  static s32   chunkSize=0;
  s32       p, pStopIndex, residue, k;
  short int  numFound, numChunks, c;
  char       plog, ratVal, entry;
  int        avgNorm, logM;
  s32       i, r, g, a, a0, a1, sk;
  int        evenB=0;
  int        rfb_log_ff, afb_log_ff;


  if (!(initialized)) {
    chunkSize = cacheSize/sizeof(short);
    mpz_init(bm); mpz_init(t);
    if (!(tmp = (short *)malloc((chunkSize+1)*sizeof(short)))) {
      fprintf(stderr, "clSieve() : Memory allocation error!\n");
      return -1;
    }
    initialized=1;
  }
  if (cacheSize/sizeof(short) != chunkSize) {
    chunkSize = cacheSize/sizeof(short);
    tmp = realloc(tmp, chunkSize*sizeof(short));
  }
  if (b%2==0) evenB=1;

  numChunks = (_a1 - _a0 + 1)/chunkSize;
  if ((_a1 - _a0 + 1)%chunkSize)
    numChunks++;
  a0 = _a0;
  a1 = _a0 +chunkSize-1;
  mpz_mul_si(bm, FB->m, b);
  logM = fplog_mpz(FB->m, FB->log_rlb);
  numFound=0;
  tStop = tmp + chunkSize;

  rfb_log_ff = (int)(FB->rfb_lambda * fplog(FB->rfb[FB->rfb_size-1], FB->log_rlb));
  afb_log_ff = (int)(FB->afb_lambda * fplog(FB->afb[FB->afb_size-1], FB->log_alb));
  for (c=0; c<numChunks; c++) {
    /* Initialize the sieve with logs */
    ratVal = (short)(-logM - fplog(b, FB->log_rlb) + rfb_log_ff);
    for (i=0; i<chunkSize; i++)
      tmp[i] = ratVal;

    pStopIndex = 2*FB->rfb_size;

    sk = SKIP_R_PRIMES;
    if (evenB) {
      sk = MAX(sk, 1); /* Skip p=2. */
      for (i=2*sk; i<pStopIndex; i+=2) {
        p = FB->rfb[i];
        plog = FB->rfb_log[i/2];
        /********************************************/
        /* Find the smallest (a, b) pair s.t.  a-bm */
        /* is divisible by 'p'.                     */
        /********************************************/
        MULMOD32(residue, FB->rfb[i+1], b, p);
        k = (residue + p - (a0 % p))%p; /* Make sure 'k' is positive. */
        if ((a0+k)%2==0) k += p;
        p += p;
        t2 = tmp+k;
        while (t2 < tStop) {
          *t2 += plog;
           t2 += p;
        }
      }
      /***** Prepare for the algebraic sieve. *****/
      /* This seems reasonable, with a -4 for luck. :) */
      avgNorm = (fplog_evalF(a0/2, b, FB) + fplog_evalF(a1/2,b,FB))/2 - 5;
      entry = (short)(afb_log_ff - avgNorm);
      for (i=0; i<chunkSize; i++) {
        tmp[i] =  (tmp[i] >= 0) ? entry : VERY_SMALL_SHORT;
      }
      pStopIndex = 2*FB->afb_size;
      sk = SKIP_A_PRIMES;
      while (FB->afb[2*sk]==2) sk++;
      for (i=2*sk; i<pStopIndex; i+=2) {
        p = FB->afb[i];
        r = FB->afb[i+1];
        MULMOD32(residue, b, r, p);
        plog = (FB->afb_log[i/2]);
        /********************************************/
        /* Find the smallest (a, b) pair s.t.  a-br */
        /* is divisible by 'p'.                     */
        /********************************************/
        k = (p + residue - (a0%p))%p;
        if ((a0+k)%2==0) k += p;
        p += p;
        t2 = tmp+k;
        while (t2 < tStop) {
          *t2 += plog;
           t2 += p;
        }
      }
    } else { /* b is odd. */
      sk = MAX(sk, 1); /* Skip p=2. */
      for (i=2*sk; i<pStopIndex; i+=2) {
        p = FB->rfb[i];
        plog = FB->rfb_log[i/2];
        /********************************************/
        /* Find the smallest (a, b) pair s.t.  a-bm */
        /* is divisible by 'p'.                     */
        /********************************************/
        MULMOD32(residue, FB->rfb[i+1], b, p);
        k = (residue + p - (a0 % p))%p; /* Make sure 'k' is positive. */
        t2 = tmp+k;
        while (t2 < tStop) {
          *t2 += plog;
           t2 += p;
        }
      }
      /***** Prepare for the algebraic sieve. *****/
      /* This seems reasonable, with a -4 for luck. :) */
      avgNorm = (fplog_evalF(a0/2, b, FB) + fplog_evalF(a1/2,b,FB))/2 - 4;
      entry = (short)(afb_log_ff - avgNorm);
      for (i=0; i<chunkSize; i++) {
        tmp[i] =  (tmp[i] >= 0) ? entry : VERY_SMALL_SHORT;
      }
      pStopIndex = 2*FB->afb_size;
      sk = SKIP_A_PRIMES;
      for (i=2*sk; i<pStopIndex; i+=2) {
        p = FB->afb[i];
        r = FB->afb[i+1];
        MULMOD32(residue, b, r, p);
        plog = (FB->afb_log[i/2]);
        /********************************************/
        /* Find the smallest (a, b) pair s.t.  a-br */
        /* is divisible by 'p'.                     */
        /********************************************/
        k = (p + residue - (a0%p))%p;
        t2 = tmp+k;
        while (t2 < tStop) {
          *t2 += plog;
           t2 += p;
        }
      }
    }

    t2=tmp;
    while (t2 < tStop) {
      if (*t2 >= 0) {
        a = (a0+ (t2-tmp));
        if ((*t2 + avgNorm - fplog_evalF(a,b,FB)) >= 0) {
          g = gcd(a, b);
          if (g*g==1)
            if (numFound < maxSmooth)
              smooth_a[numFound++] = a;
        } 
      }
      t2++;
    }
    /* Update the interval for the next pass: */
    a0 += chunkSize;
    a1 += chunkSize;
  }

  return numFound;
}

typedef struct {
  s32 p;
  s32 k;
  char plog;
  char blank0;
  short blank1;
  s32  blank2; /* These two pad it out to 16 bytes apiece. */
} local_fb_t;


/*******************************************************************************/
int clCacheFriendlySieve(s32 *smooth_a, int maxSmooth, nfs_fb_t *FB, 
                         s32 _a0, s32 _a1, s32 b, s32 cacheSize)
/*******************************************************************************/
/* Sieve from (a0, b) to (a1, b). Put the smooth a's in smooth_a.              */
/* That is, sieve on a - b*m, and store the values of 'a' that                 */
/* give smooth values.                                                         */
/* Return value: Number of relations found, or < 0 on error.                   */
/*******************************************************************************/
/* We only pay attention to cacheSize the first time this function is called - */
/* so be sure to get it right the first time!!!!                               */
/*******************************************************************************/
/* The idea is the following: allocate a chunk of memory that will fit in cache*/
/* and use it as follows:                                                      */
/*  -------------------------------------------------------------------------  */
/*  | a few primes |                   sieve array                          |  */
/*  -------------------------------------------------------------------------  */
/* We also keep a local copy of both factor bases, so that we can modify       */
/* it between chunks, so as to avoid re-computing the same things over and     */
/* over. While sieving, we grab some primes from the local copy, stick them    */
/* in the cache space, then sieve with them. When done, they are prepared for  */
/* the next chunk, and re-written to the local fb space.                       */
/*******************************************************************************/
/* I tried the above, and it gave no improvement on my Athlon XP 2800. So this */
/* does not actually do that. But it does make a local copy of the primes which*/
/* it uses to remember the state from one piece to the next.                   */
/*******************************************************************************/
{ static int    initialized=0;
  static mpz_t  bm, t;
  static char  *sieveSpace;
  static s32   chunkSize=0;
  static s32   rfbUseSize=0, afbUseSize=0;
  static local_fb_t   *fbSpace=NULL;
  char register *t2, *tStop;
  s32          p, residue, k, r;
  short int     numFound, numChunks, c, tmpInt, pInftysLog;
  char          plog, ratVal, entry;
  int           avgNorm, logM;
  s32          i, j, g, a, a0, a1, skR, skA;
  int           evenB=0; 
  int           rfb_log_ff, afb_log_ff;

  rfbUseSize = FB->rfb_size;
  afbUseSize = FB->afb_size;
  if (!(initialized)) {
    chunkSize = cacheSize/sizeof(char) - 1 ;
    mpz_init(bm); mpz_init(t);
    if (!(sieveSpace = (char *)malloc(cacheSize))) {
      fprintf(stderr, "clSieve() : Memory allocation error!\n");
      return -1;
    }
    if (!(fbSpace = (local_fb_t *)malloc(sizeof(local_fb_t)*
                             (FB->rfb_size + FB->afb_size)))){
      fprintf(stderr, "clSieve() : Memory allocation error!\n");
      return -1;
    }
    initialized=1;
  }
  if (b%2==0) evenB=1;

  pInftysLog = 0;
  numChunks = (_a1 - _a0 + 1)/chunkSize;
  if ((_a1 - _a0 + 1)%chunkSize)
    numChunks++;
  a0 = _a0;
  a1 = _a0 +chunkSize-1;
  mpz_mul_si(bm, FB->m, b);
  logM = fplog_mpz(FB->m, FB->log_rlb);
  numFound=0;

/* Really, we could reuse this stuff; we could first check to see
   what the last 'b' value was that we sieved with. It would then
    be easy to update for the next b and sieve in the opposite
   direction. But I think that this will have almost no effect
   on the runtime, so I will wait to try.
*/
 
  rfb_log_ff = (int)(FB->rfb_lambda * fplog(FB->rfb[FB->rfb_size-1], FB->log_rlb));
  afb_log_ff = (int)(FB->afb_lambda * fplog(FB->afb[FB->afb_size-1], FB->log_alb));

  /*********************************************/
  /* Prepare the local FB for the first chunk. */
  /*********************************************/
  skR = SKIP_R_PRIMES;
  if (evenB) skR = MAX(skR, 1); /* Skip p=2. */
  for (i=0; i<FB->rfb_size; i++) {
    p = FB->rfb[2*i];
    plog = FB->rfb_log[i];
    /********************************************/
    /* Find the smallest (a, b) pair s.t.  a-bm */
    /* is divisible by 'p'.                     */
    /********************************************/
    MULMOD32(residue, FB->rfb[2*i+1], b, p);
    k = (residue + p - (a0 % p))%p; /* Make sure 'k' is positive. */
    if (evenB) {
      if ((a0+k)%2==0) k += p;
      p += p;
    }
    fbSpace[i].p=p; 
    fbSpace[i].k = k;
    fbSpace[i].plog = plog;
  }
  skA = SKIP_A_PRIMES;
  if (evenB) /* Skip p=2. */
    while (FB->afb[2*skA]==2) skA++;
  for (i=j=0; i<FB->afb_size; i++) {
    p = FB->afb[2*i];
    r = FB->afb[2*i+1];
    while (r==p) {
      if (b%p == 0) 
        pInftysLog += FB->afb_log[i];
      
      afbUseSize--;
      i++;
      p = FB->afb[2*i];
      r = FB->afb[2*i+1];
    }   
    plog = FB->afb_log[i];
    MULMOD32(residue, r, b, p);
    k = (residue + p - (a0 % p))%p; /* Make sure 'k' is positive. */
    if (evenB) {
      if ((a0+k)%2==0) k += p;
      p += p;
    }
    fbSpace[j+rfbUseSize].p = p; 
    fbSpace[j+rfbUseSize].k = k; 
    fbSpace[j+rfbUseSize].plog = plog;
    j++;
  }
  tStop = sieveSpace + chunkSize;
  tmpInt = (short)(-logM - fplog(b, FB->log_rlb) + rfb_log_ff);
  ratVal = ((tmpInt >= -127) ? (char)tmpInt : -127);


  /**************************************/
  /* Ok - let's rock!                   */
  /**************************************/
  for (c=0; c<numChunks; c++) {
    if (forceStop) goto CLS_CLEANUP;
    for (i=chunkSize-1; i>0; i--)
      sieveSpace[i] = ratVal;
    sieveSpace[i]=ratVal;
    j = skR;
    while (j < rfbUseSize) {
      p = fbSpace[j].p;
      t2 = sieveSpace + fbSpace[j].k;
      plog = fbSpace[j].plog;
      while (t2 < tStop) {
        *t2 += plog;
         t2 += p;
      }
      fbSpace[j].k = (s32)(t2-tStop);
      j++;
    }

    /***** Prepare for the algebraic sieve. *****/
    /* This seems reasonable, with a -4 for luck. :) */
#ifdef _OLD
    avgNorm = (fplog_evalF(a0/2, b, FB) + fplog_evalF(a1/2,b,FB))/2 - 4;
    tmpInt = (short)(afb_log_ff - avgNorm + pInftysLog);
    entry = (tmpInt >= -127) ? tmpInt : -127;
    for (i=0; i<chunkSize; i++) {
      sieveSpace[i] =  (sieveSpace[i] >= 0) ? entry : VERY_SMALL_SHORT;
    }
#else
    i=0; 
    while (i<chunkSize) {
      j = MIN(chunkSize-i, 1024);
      avgNorm = fplog_evalF(a0+i, b, FB) - 2;
      tmpInt = (short)(afb_log_ff - avgNorm + pInftysLog);
      entry = ((tmpInt >= -127) ? tmpInt : -127);
      for ( ; j> 0; j--) {
        sieveSpace[i] =  (sieveSpace[i] >= 0) ? entry : -127;
        i++;
      }
    }
#endif


    i=0;

    j = rfbUseSize+skA;
    while (j < rfbUseSize+afbUseSize) {
      p = fbSpace[j].p;
      t2 = sieveSpace + fbSpace[j].k;
      plog = fbSpace[j].plog;
      while (t2 < tStop) {
        *t2 += plog;
         t2 += p;
      }
      fbSpace[j].k = (s32)(t2-tStop);
      j++;
    }

    /* Look for candidates. */
    t2=sieveSpace;
    while (t2 < tStop) {
      if (*t2 >= 0) {
        a = (a0+ (t2-sieveSpace));
        g = gcd(a, b);
        if (g*g==1)
          if (numFound < maxSmooth)
            smooth_a[numFound++] = a;
      }
      t2++;
    }
    /* Update the interval for the next pass: */
    a0 += chunkSize;
    a1 += chunkSize;
  }
CLS_CLEANUP:
  return numFound;
}


  
